use crate::{
    decrypt_file_buffered,
    errors::{CvfsErrors, CvfsResult},
    metadata::{FileTag, Metadata},
    open_rwlocked,
};
use multi_key_manager::{file::File, LiveKeyMaterial, OpeningNonce};
use ring::aead::OpeningKey;
use std::{
    io::{Read, Write},
    path::Path,
    sync::{Mutex, Weak},
    thread::panicking,
};

/// an open active file
pub struct ActiveFile {
    /// Clonable needed for metadata
    pub(crate) id: FileTag,
    /// The actual data buffer
    data: Vec<u8>,
    /// seekable
    offset: usize,
    /// Locked file
    real_file: File,
}
impl ActiveFile {
    pub fn open_rwlocked(
        real_path: &Path,
        id: FileTag,
        file_key: &mut OpeningKey<OpeningNonce>,
    ) -> CvfsResult<Self> {
        let mut real_file = open_rwlocked(real_path)?;

        let data = if real_file.empty()? {
            Vec::new()
        } else {
            decrypt_file_buffered(&mut real_file, file_key)
                .map_err(|_generic_ring_error| CvfsErrors::LeafDecryptionError)?
        };

        Ok(Self {
            id,
            data,
            offset: 0,
            real_file,
        })
    }
}
impl Drop for ActiveFile {
    fn drop(&mut self) {
        if let Err(e) = self.flush() {
            let error = format!("unable to flush aead_vfs::ActiveFile: {:?}", e);
            if panicking() {
                eprintln!("{error}");
            } else {
                panic!("{error}");
            }
        }
    }
}
impl Write for ActiveFile {
    /// All writes are to the buffer and not commited to disk until `.flush()?` is called or the file closes
    fn write(&mut self, buffer: &[u8]) -> Result<usize, std::io::Error> {
        let bytes_written = buffer.len();
        let new_max_len = self.offset + buffer.len();
        if self.data.len() < new_max_len {
            self.data.resize_with(new_max_len, Default::default);
        }
        let target = &mut self.data[self.offset..buffer.len()];
        target.copy_from_slice(buffer);

        Ok(bytes_written)
    }
    /// Do the encryption
    fn flush(&mut self) -> Result<(), std::io::Error> {
        todo!("ActiveFile::flush")
    }
}

/// These Handlers should reduce the time the main Mutex is held so concurrent access is possible
pub struct SecureFileHandler {
    pub(crate) metadata: Weak<Mutex<Metadata>>,
    pub(crate) id: FileTag,
    /// the current seek position
    pub(crate) offset: usize,
}

#[cfg(test)]
mod tests {
    use super::*;
    use ring::aead::{BoundKey, UnboundKey};
    use std::path::PathBuf;

    /// ext must not start with a `/`
    fn path<P: AsRef<Path>>(ext: P) -> PathBuf {
        let mut p = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        // we are in a cargo workspace
        p.pop();
        p.push("target");
        p.push("aead-vfs_tests-secure-file");
        std::fs::create_dir_all(&p).expect("unable to create testing directory");
        p.push(ext);
        p
    }

    fn test_key() -> OpeningKey<OpeningNonce> {
        let key_material = [42u8; 32];
        let nonce = OpeningNonce::new(1);
        let algorithm = &ring::aead::CHACHA20_POLY1305;

        let unbound_key =
            UnboundKey::new(algorithm, &key_material).expect("unable to construct UnboundKey");
        OpeningKey::new(unbound_key, nonce)
    }
    #[test]
    fn simple_write() {
        let mut key = test_key();
        let mut a = ActiveFile::open_rwlocked(&path("simple_write.bin"), FileTag(1), &mut key)
            .expect("unable to open test file");

        a.write(b"abc").unwrap();
        println!("a.data: {:?}", a.data);
        assert_eq!(3, a.data.len());
        assert_eq!(3, a.offset);
        /*
            a.write(b"abc");
            assert_eq!(6, a.offset);
            assert_eq!(6, a.data.len());
        */
    }
}
